noflo = require 'noflo'
log = require 'graceful-logger'
messaging = require './messaging'

loader = null

exports.getLoader = ->
  loader = new noflo.ComponentLoader __dirname unless loader
  loader

exports.wrapInport = (instance, port, channel, queue) ->
  preprocess = (ins, callback) -> callback ins
  if typeof queue is 'object'
    preprocess = queue.preprocess
    queue = queue.queue

  socket = noflo.internalSocket.createSocket()
  instance.inPorts[port].attach socket
  channel.consume queue, (msg) ->
    return unless msg
    message = JSON.parse msg.content.toString()

    repo = message.site or message.repo or message.type or ''
    siteUrl = message.url or message.path or message.cover?.src or ''
    log.info "#{queue.toUpperCase()} -> #{msg.fields.deliveryTag} #{repo} #{siteUrl}"

    preprocess message, (processed) ->
      socket.beginGroup msg.fields.deliveryTag
      socket.send processed
      socket.endGroup()
      socket.disconnect()

exports.wrapOutport = (instance, port, channel, queue) ->
  socket = noflo.internalSocket.createSocket()
  instance.outPorts[port].attach socket
  groups = []
  handled = []
  socket.on 'begingroup', (group) ->
    groups.push group
  socket.on 'data', (data) ->
    grps = groups.slice 0
    if typeof queue is 'function'
      queue data, grps, (err, retVal) ->
        repo = data.site or data.repo or data.type or ''
        siteUrl = data.url or data.path or data.cover?.src or ''
        if err
          log.err "#{instance.nodeId} NACK #{grps[0]} #{repo} #{siteUrl}"
          log.err err
          operation = 'nack'
        else
          log.info "#{instance.nodeId} ACK #{grps[0]} #{repo} #{siteUrl}"
          operation = 'ack'
        if grps.length and typeof grps[0] is 'number'
          return unless handled.indexOf(grps[0]) is -1
          channel[operation]
            fields:
              deliveryTag: grps[0]
          handled.push grps[0]
      return

    # Data received, ACK/NACK original packet
    operation = 'ack'
    operation = 'nack' if queue is 'nack'
    if grps.length and typeof grps[0] is 'number' and handled.indexOf(grps[0]) is -1
      channel[operation]
        fields:
          deliveryTag: grps[0]
      handled.push grps[0]

    repo = data.site or data.repo or data.type or ''
    siteUrl = data.url or data.path or data.cover?.src or ''

    if queue is 'ack'
      log.info "#{instance.nodeId} #{queue.toUpperCase()} #{grps[0]} #{repo} #{siteUrl}"
      return
    if queue is 'nack'
      log.err "#{instance.nodeId} #{queue.toUpperCase()} #{grps[0]} #{repo} #{siteUrl}"
      log.err data
      return

    # Send to outport
    channel.sendToQueue queue, new Buffer JSON.stringify data
    log.info "#{queue.toUpperCase()} <- #{grps[0]} #{repo} #{siteUrl}"

  socket.on 'endgroup', (group) ->
    groups.pop()
  socket.on 'disconnect', ->
    groups = []

exports.wrap = (graph, inQueues, outQueues, callback) ->
  loader = exports.getLoader()

  qs = []
  qs.push val for key, val of inQueues
  qs.push val for key, val of outQueues

  queues = qs.map((q) ->
    return q.queue if typeof q is 'object'
    q
  ).filter (q) ->
    return false if typeof q is 'function'
    return false if q is 'ack' or q is 'nack'
    true

  messaging.prepareQueues queues, null, (err, channel) ->
    return callback err if err
    # Allow only one message per worker before ack/nack
    channel.prefetch 1

    loader.load graph, (err, instance) ->
      return callback err if err

      [library, graphName] = graph.split '/'
      instance.nodeId = graphName.toUpperCase()
      if instance.isReady()
        instance.start() if instance.network
        for port, queue of inQueues
          exports.wrapInport instance, port, channel, queue
        for port, queue of outQueues
          exports.wrapOutport instance, port, channel, queue
        return callback null, instance
      instance.once 'ready', ->
        instance.start() if instance.network
        for port, queue of inQueues
          exports.wrapInport instance, port, channel, queue
        for port, queue of outQueues
          exports.wrapOutport instance, port, channel, queue
        callback null, instance

exports.showFailStats = (failures) ->
  unless failures.length
    console.log 'No solutions, and no failure log available'
    return

  closest = null
  reasons = {}
  remaining = 0

  failures.forEach (failure) ->
    closest = failure unless closest
    if failure.remaining < closest.remaining
      closest = failure
    if failure.remaining is 0 and failure.reason is 'cannot_be_tail'
      closest = failure
    reasons[failure.reason] = 0 unless reasons[failure.reason]
    reasons[failure.reason]++
    remaining += failure.remaining.length

  averageRemaining = Math.floor remaining / failures.length
  console.log "Solution branches attempted: #{failures.length}"
  console.log "Average blocks remaining: #{averageRemaining}"
  console.log "Reasons to fail:"
  for reason, howMany of reasons
    console.log " * #{reason}: #{howMany}"
  console.log ""
  console.log "Closest solution failed because #{closest.reason}"
  if closest.solution.templates.length is 0
    console.log "No templates used successfully in solution"
  else
    console.log "Templates used", closest.solution.templates.map((template) -> template.templateId).join ', '
    last = closest.solution.templates[closest.solution.templates.length - 1]
    if last.content
      console.log "Last template contained:"
      for type, blocks of last.content
        console.log " * #{type}: #{blocks.length}"
      if last.template.rules?.consumes?.length
        console.log "Last template could have still consumed"
        for consumer in last.template.rules.consumes
          console.log " * #{consumer.type}: #{consumer.number.min - consumer.number.now}"
        if closest.lastBlock
          console.log "Instead we had #{closest.lastBlock.type}"
  console.log ""
  if closest.remaining.length is 0
    console.log 'There were no items remaining'
  else
    console.log "#{closest.remaining.length} remaining items started with", closest.remaining.splice(0, 5).map((item) -> item.content.map((block) -> block.type)).join ', '
